/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 */
#ifndef BINARY_DATA_COMMON_H
#define BINARY_DATA_COMMON_H

#include <stdio.h>
#include <string>
#include "UtilNumberType.h"
#include "BinaryUdfFile.h"

namespace binaryio {

// CONSTANT
// STRING
#define BINARY_MAX_STRING_SIZE 256
#define BINARY_READ_STRING_SIZE (BINARY_MAX_STRING_SIZE-1)
// VERSION
#define BINATY_FORMAT_VERSION_MAJOR   1
#define BINATY_FORMAT_VERSION_MINOR   0
#define BINATY_FORMAT_VERSION_MAINTE  0
#define BINATY_FORMAT_VERSION_DELIMIT '.'
#define BINATY_FORMAT_VERSION         "%02d.%02d.%02d"
#define BINATY_FORMAT_VERSION_LEN     8
// ENDIAN
#define BINATY_BIG_ENDIAN_KEY		'B'
#define BINATY_LITTLE_ENDIAN_KEY	'L'
// KEY WORDS
#define BINARY_DATA_KEY_NUMBER	6
#define BINARY_DATA_KEY_LENGTH	4
#define BINARY_DATA_START_KEY	"DATA"
#define BINARY_DATA_NAMEP_KEY	"NPIF"
#define BINARY_DATA_STEP0_KEY	"COMN"
#define BINARY_DATA_STEPX_KEY	"STEP"
#define BINARY_DATA_DIVID_KEY	"DIVD"
#define BINARY_DATA_BEGIN_KEY	"BGND"
#define BINARY_DATA_START_KEYID	0
#define BINARY_DATA_NAMEP_KEYID 1
#define BINARY_DATA_STEP0_KEYID	2
#define BINARY_DATA_STEPX_KEYID	3
#define BINARY_DATA_BEGIN_KEYID	4
#define BINARY_DATA_DIVID_KEYID	5
#define BINARY_DATA_EOF_KEYID	99
// DIVIDED RECORD KEY WORDS
#define BINARY_DIVID_DIRECTORY	"RECORDDIR:"
#define BINARY_DIVID_DIRECTORY_LENGTH 10
#define BINARY_DIVID_RECORD		"RECORD:"
#define BINARY_DIVID_RECORD_LENGTH 7
#define BINARY_DIVID_OBJECT		"OBJECT:"
#define BINARY_DIVID_OBJECT_LENGTH 7
#define BINARY_DIVID_OBJECTDIR	"OBJECTDIR:"
#define BINARY_DIVID_OBJECTDIR_LENGTH 10

// TYPE
typedef float  REAL32TYPE;
typedef double REAL64TYPE;

// Functions.
// Returns byte size.
unsigned int lengthBinarySTRINGL8( const char* text );
unsigned int lengthBinarySTRINGL16( unsigned int len );
unsigned int lengthBinarySTRINGL16( const char* text );
unsigned int lengthBinarySTRINGNL( unsigned int len );
unsigned int lengthBinarySTRINGNL( const char* text );
bool copyBinarySTRINGNL( FILE* destination, FILE* source, unsigned int* optCopiedSize = 0);
int skipBinarySTRINGNL( FILE* binaryOpened );

// Checker.
#define BINARY_DATA_CHECKER_SIZE 8
#define BINARY_DATA_CHECKER_INDX 6
#define BINARY_DATA_CHECKER_CODE (BINARY_DATA_CHECKER_SIZE+BINARY_DATA_CHECKER_INDX)
void copyBinaryChecker( unsigned char* bits );
unsigned char* getBinaryChecker();
unsigned char checkBinaryChecker( const unsigned char* bits );
bool checkBinaryBit( const unsigned char bit );
// Character fwrite/fread.
size_t fwrite2( const char* buffer, size_t size, size_t count, FILE *stream, crypt_udf* crypt);
size_t fread2( char* buffer, size_t size, size_t count, FILE *stream, crypt_udf* crypt);

// Binary swap for the different endian platforms.
class BinaryInterface {
private:
	bool _needSwap;
	crypt_udf* _crypt;
public:
	BinaryInterface():_needSwap(false),_crypt(0){}
	~BinaryInterface(){
		if(_crypt) delete _crypt;
	}
	void clear(){ _needSwap = false; }
	void swap(){ _needSwap = true; }
	////////////////////////////////////
	// According to the endian, the bytes of value is swapped.
	void eswap( short& value );
	void eswap( unsigned short& value );
	void eswap( int& value );
	void eswap( unsigned int& value );
	void eswap( longint_t& value );
	void eswap( ulongint_t& value );
	void eswap( float& value );
	void eswap( double& value );
	void eneswap( double& value );
	void deeswap( double& value );
	//// writer ////
	static bool writeUCHAR( unsigned char number, FILE* binaryOpened );
	static bool writeUCHARARRAY( unsigned int number, unsigned char* array, FILE* binaryOpened );
	bool writeINT16( short number,  FILE* binaryOpened );
	bool writeUINT16( unsigned short number,  FILE* binaryOpened );
	bool writeINT32( int number,  FILE* binaryOpened );
	bool writeUINT32( unsigned int number,  FILE* binaryOpened );
	bool writeINT64( longint_t number,  FILE* binaryOpened );
	bool writeUINT64( ulongint_t number,  FILE* binaryOpened );
	bool writeREAL32( float number, FILE* binaryOpened );
	bool writeREAL64( double number, FILE* binaryOpened );
	bool writeSTRINGL8( const char* text, FILE* binaryOpened );
	bool writeSTRINGL16( const char* text, FILE* binaryOpened );
	bool writeSTRINGL32( const char* text, FILE* binaryOpened );
	bool writeSTRINGL32( const std::string& text, FILE* binaryOpened);
	bool writeSTRINGNL(
		const char* text,
		FILE* binaryOpened
	);
	static bool writeSTRINGNL(
		const char* text,
		FILE* binaryOpened,
		crypt_udf* crypt
	);
	static bool writeSTRINGNL(
		const unsigned int length,
		const char* text,
		FILE* binaryOpened,
		crypt_udf* crypt = NULL
	);
	static bool writeSTRINGNL(
		const std::string& text,
		FILE* binaryOpened
	);
	static bool writeSTRING_NOLENGTH(
		const char* text,
		FILE* binaryOpened
	);
	//// Reader ////
	static bool readUCHAR( unsigned char& number, FILE* binaryOpened );
	static bool readUCHARARRAY( unsigned int number, unsigned char* array, FILE* binaryOpened );
	bool readINT16( short& number,  FILE* binaryOpened );
	bool readUINT16( unsigned short& number,  FILE* binaryOpened );
	bool readINT32( int& number,  FILE* binaryOpened );
	bool readUINT32( unsigned int& number,  FILE* binaryOpened );
	bool readINT64( longint_t& number,  FILE* binaryOpened );
	bool readUINT64( ulongint_t& number,  FILE* binaryOpened );
	bool readREAL32( float& number, FILE* binaryOpened );
	bool readREAL64( double& number, FILE* binaryOpened );
	static bool readSTRINGL8( char** text, FILE* binaryOpened ); // new char[N];
	bool readSTRINGL32( char** text, FILE* binaryOpened );
	bool readSTRINGL32( std::string& text, FILE* binaryOpened );
	// bool readSTRINGL16( char** text, FILE* binaryOpened );
	// bool readSTRINGNL( char** text, FILE* binaryOpened );
	static bool readSTRINGNL( std::string& text, FILE* binaryOpened );
	//// Crypt ////
	void setupCrypt( int code );
	crypt_udf* getCrypt();
	bool isCrypt();
	//// Skip ////
	static bool skipUCHAR( FILE* binaryOpened );
	static bool skipINT16( FILE* binaryOpened );
	static bool skipUINT16( FILE* binaryOpened );
	static bool skipINT32( FILE* binaryOpened );
	static bool skipUINT32( FILE* binaryOpened );
	static bool skipINT64( FILE* binaryOpened );
	static bool skipUINT64( FILE* binaryOpened );
	static bool skipREAL32( FILE* binaryOpened );
	static bool skipREAL64( FILE* binaryOpened );
	static void skipBYTE( FILE* binaryOpened, int byte, size_t number );
};


} // end of namespace.
#endif
