/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 */
#ifndef BINARY_DATA_READ_H
#define BINARY_DATA_READ_H
#include "StructureCacheComponent.h"
#include "BinaryDataInfo.h"
#include "UtilNumberType.h"
#include "recycle_string.h"
#include "DataPositionType.h"
#include <string>
#include <vector>
using namespace CacheUtilCore;

namespace binaryio {
#define BINARY_READ_DIVIDE	(-9)
#define BINARY_READ_DIVIDE_END	(-19)

struct SeriesPair {
	int	id;
	longint_t position;
};
typedef std::vector<SeriesPair> SeriesPositions;

class BinarySeriesPosition {
private:
	std::vector<SeriesPositions> _sections;
public:
	BinarySeriesPosition(){}
	~BinarySeriesPosition(){}
	void createDimension(int dimSection){
		_sections.resize(dimSection);
	}
	void clearSeriesPosition( int section = -999 )
	{
		if(section < 0){
			for(unsigned int i=0;i<_sections.size();i++){
				_sections[i].clear();
			}
		}
		else{
			if(section < (int)_sections.size()){
				_sections[section].clear();
			}
		}
	}
	SeriesPositions& getSeriesPosition( int section ){
		if(section < 0 || section >= (int)_sections.size()) throw 0;
		return _sections[section];
	}
};

class BinaryDataRead {
private:
	FILE* _binaryOpened;
	BinaryDataInfo* _binaryInfo;
	BinaryNameManager* _nameManager;
	StructureCacheComponent* _baseComponent;
	SortedDataPositionList* _data_position_list;
	/////////////////
	UserBinaryUdfData* _user_data;
	////////////////////////////////////////
	BinaryInterface _binaryIO;
	char _temporary[BINARY_MAX_STRING_SIZE];
	recycle_string _buffer;
	////////////////////////////////////////
	bool switch_partial_read_array;
	unsigned int limit_size_of_read_array;
	std::string _bin_error_message;

private:
	bool checkBits( const unsigned char* bits );
	int readBinarySTRINGL8( char** text, FILE* binaryOpened );
	int readBinarySTRINGL16( char** text, FILE* binaryOpened );
	int readBinarySTRINGNL( char** text, FILE* binaryOpened );
	int skipBinarySTRINGNL( FILE* binaryOpened );
	int skipBinarySTRINGNL( unsigned int skip, FILE* binaryOpened );
	int skipArray(unsigned int skip, unsigned int size, FILE* binaryOpened);
	void setDataPosition(unsigned int id, ulongint_t position);
	void resetAllDataPosition();

public:
	// Constructor/Destructor
	BinaryDataRead(
		FILE* binaryOpened,
		BinaryDataInfo* binaryInfo,
		BinaryNameManager* nameManager,
		StructureCacheComponent* baseComponent,
		SortedDataPositionList* data_position_list = 0
	);
	~BinaryDataRead();
	///// Public methods //////////////////////////////////////
	static const char* getKey( int id );
	static int searchKey( FILE* binaryOpened, int* shiftFineTuning = 0 );
	bool readVersion();
	bool checkBinaryByte();
	bool readEndian();
	bool readInfo( BinaryDataInfo* info, unsigned short* datasize = 0 );
	bool readDataNameID( bool readKey = true );
	bool readStep(
		int&             section,
		std::string&     label,
		longint_t&       datasize,
		int*             shiftFineTuning
	);
	bool readSeries(
		int&             section,
		std::string&     label,
		longint_t&       datasize
	);
	/**
	 * @brief Skip a binary step data.
	 * @brief Read the header data and skip the data.
	 * @param (section) return 0:common, 1:record
	 * @param (label) return record label
	 * @param (datasize) return step data size
	 * @param (shiftFineTuning) gap of keyword location
	 * @return true for normal end
	 */
	bool skipStep(
		int&             section,
		std::string&     label,
		longint_t&       datasize,
		int*             shiftFineTuning
	);
	SortedDataPositionList* getDataPositionList(){
		return _data_position_list;
	}
	void setDataPositionList(SortedDataPositionList* data_pos_list){
		_data_position_list = data_pos_list;
	}
	/////////////////////////////////////////////////////////////
	void setErrorMessage( int eid, const char* name = 0 );
	void setErrorMessage( int eid, const int number );
	std::string& getErrorMessage(){
		return _bin_error_message;
	}
	/**
	 * @fn
	 * Get the limit size of reading array data.
	 */
	/**
	 * @brief Get the limited size of reading array data
	 * @return limited size of reading array data
	 */
	unsigned int getLimitSizeOfArray(){
		return limit_size_of_read_array;
	}
	/**
	 * @brief Return whether partial reading array is effective
	 * @return true if partial reading array is effective
	 */
	bool isPartialReadArray(){
		return switch_partial_read_array;
	}
	/**
	 * @brief Reread data of a variable
	 * @param[in] (id) data definition id
	 * @param[in] (recordPos) record start position
	 * @return true if the work has done
	 */
	bool rereadData(int id, longint_t recordPos);
	/**
	 * @brief Reread all data in a record
	 * @param[in] (recordPos) record start position
	 * @return true if the work has done
	 */
	bool rereadAllData(longint_t recordPos);
	/**
	 * @brief Read divide key and the following data.
	 * @param[out] (shiftFineTuning) position fine tuning
	 * @return true if key is 'DIVD'.
	 */
	int readBinDivideKey(int* shiftFineTuning = 0);
	/**
	 * @brief Get optional divided binary udf data.
	 * @return UserBinaryUdfData
	 */
	UserBinaryUdfData* getUserBinaryUdfData() {
		return _user_data;
	}

private:
	// READ.
	bool readStepHeader(
		int&              section,
		std::string&      label,
		unsigned int&     datasize,
		int*              shiftFineTuning,
		SeriesPositions*  series = 0
	);
	bool readDivideHeader();
	bool readSeriesInfo( unsigned int& datasize, SeriesPositions& series );
	bool readStepData( int section, longint_t& datasize );
	bool readSections( int section );
	bool readSection(
		unsigned char typeID,
		int section,
		StructureCacheComponent* component
	);
	bool readValue(
		unsigned int number,
		unsigned char typeID,
		BaseDataType baseType,
		CacheSection* csection
	);
	// Skip.
	// SKIP.
	bool skipStepHeader(
		int&	section,
		int*	shiftFineTuning = 0
	);
	bool skipStepData( longint_t& datasize );
	bool skipSeriesInfo( unsigned int& datasize );
	bool skipSection(
		unsigned char typeID,
		unsigned char partArrayDim,
		unsigned short fullArrayDim
	);
	bool skipValue(
		unsigned int number,
		unsigned char typeID,
		unsigned char typeByte,
		unsigned short fullArrayDim
	);
	bool skipArraySize(unsigned char adim);
	// Reread
	int rereadStepDataByPosition( int find_id, int section );
	bool rereadAllStepData( int section );
	bool rereadFindStepData( int find_id, int section );
	bool rereadSection(
		unsigned char typeID,
		int section,
		StructureCacheComponent* component
	);
	bool rereadValue(
		unsigned char typeID,
		BaseDataType baseType,
		CacheSection* csection
	);
};

} // End of namespace.
#endif
