
#include "BinaryDataRead.h"
#include "StructureCacheInterface.h"
#include "UtilHugeFile.h"
#include "CacheEnviron.h"
#include "UtilArrayEquilibration.h"
using namespace std;
using namespace CacheUtilCore;
//using namespace BinaryFiles;

namespace binaryio {

bool BinaryDataRead::rereadAllData(longint_t recordPos)
{
	if(recordPos >= 0){
		if(UtilHugeFile::seek(_binaryOpened,recordPos,SEEK_SET)){
			return false;
		}
	}
	// Read step data.
	int section = 0;
	if(!skipStepHeader(section, 0)){
		return false;
	}
	return rereadAllStepData(section);
}

bool BinaryDataRead::rereadData(int id, longint_t recordPos)
{
	if(recordPos >= 0){
		if(UtilHugeFile::seek(_binaryOpened,recordPos,SEEK_SET)){
			return false;
		}
	}
	// Read step data.
	int section = 0;
	if(!skipStepHeader(section, 0)){
		return false;
	}
	int status = rereadStepDataByPosition(id, section);
	switch(status){
	case 0:
		// No position data, so reread by FindStepData.
		break;
	case 1: // Error.
		return false;
	case 2: // Read normal.
		return true;
	case 4: // No data to read.
		return true;
	default:
		if(recordPos >= 0){
			if(UtilHugeFile::seek(_binaryOpened,recordPos,SEEK_SET)){
				return false;
			}
		} else {
			return false;
		}
		// Reread FindStepData.
		break;
	}
	return rereadFindStepData(id, section);
}

	
bool BinaryDataRead::skipStepHeader(
	int&	section,
	int*	shiftFineTuning
)
{
	unsigned int datasize = 0;
	// KEY -> section.
	int num = searchKey(_binaryOpened, shiftFineTuning);
	switch(num){
	case BINARY_DATA_STEP0_KEYID: // COMN
		section = 0;
		break;
	case BINARY_DATA_STEPX_KEYID: // STEP
		section = 1;
		break;
	case BINARY_DATA_EOF_KEYID:
		datasize = 0;
		return true;
	default:
		return false;
	}
	// label
	num = skipBinarySTRINGNL(_binaryOpened);
	// series.
	if(!skipSeriesInfo(datasize)) return false;
	// serise + label
	datasize += lengthBinarySTRINGNL((unsigned int)num);
	return true;
}

/**
 * @fn
 * Reread an array data by position data
 * @param (find_id) binary data id
 * @param (section) 0:common, 1:record
 * @return file position status (-1:uncertain, 0:no change, 1:data start, 2:data end)
 */
int BinaryDataRead::rereadStepDataByPosition( int find_id, int section )
{
	if(_data_position_list){
		DataPositionType dd(find_id);
		DataPositionType* found = _data_position_list->search(&dd);
		if(!found){
			return 0;
		}
		ulongint_t position = found->getPosition();
		if (position == 0) {
			// No data to read.
			StructureCacheComponent* component = _nameManager->findComponentFromBinaryID(find_id);
			if (!component) {
				return 1;
			}
			CacheSection* csection = component->section(section);
			if (!csection) return -1;
			LimitArraySection* arsection = dynamic_cast<LimitArraySection*>(csection);
			if (arsection) {
				arsection->setSizeOfRead(0);
			}
			return 4;
		}
		// SEEK
		if(UtilHugeFile::seek(_binaryOpened,position,SEEK_SET)){
			return -1;
		}
		const char* name = 0;
		unsigned char typeID = 0;
		unsigned char partArrayDim = 0;
		int id;
		int parentID = 0;
		if(!_nameManager->getBinaryAttribute(find_id,&name,&parentID,&typeID,&partArrayDim)){
			return 1;
		}
		// Component.
		StructureCacheComponent* component = _nameManager->findComponentFromFullName(name);
		if(!component){
			return 1;
		}
		// section.
		if(rereadSection(typeID,section,component)){
			return 2;
		}
		return -1;
	}
	return 0;
}

bool BinaryDataRead::rereadAllStepData( int section )
{
	// STEP SIZE DATA
	ulongint_t udatasize = 0;
	if(!_binaryIO.readUINT64(udatasize,_binaryOpened)) return false;
	
	// NUMBER
	unsigned int counter = 0;
	if(!_binaryIO.readUINT32(counter,_binaryOpened)) return false;
	
	//////////////////////////////////////////////////////////////
	const char* name = 0;
	unsigned char typeID = 0;
	unsigned char partArrayDim = 0;
	int id;
	int parentID = 0;
	StructureCacheComponent* component;
	for(unsigned int i=0;i<counter;i++){
		// ID
		if(!_binaryIO.readINT32(id,_binaryOpened)) return false;
		if(id > 0){
			// ParentID
			if(!_nameManager->getBinaryAttribute(id,&name,&parentID,&typeID,&partArrayDim)){
				setErrorMessage( 10 ); // Collapsed.
				return false;
			}
		}
		// Component.
		component = _nameManager->findComponentFromFullName(name);
		if(component == 0){
			return false;
		}
		unsigned short fullArrayDim = component->attribute()->fullDim();
		if(fullArrayDim == 0){
			// skip read.
			if(!skipSection(typeID, partArrayDim, fullArrayDim)) return false;
			continue;
		}
		// For array data.
		CacheSection* csection = component->section(section);
		LimitArraySection* asection = (LimitArraySection*) csection;
		if(asection->sizeOfRead() == 0){
			// No reread data. -> skip read.
			if(!skipSection(typeID, partArrayDim, fullArrayDim)) return false;
			continue;
		}

		// Reread section.
		if(!rereadSection(typeID,section,component)){
			return false;
		}
	}
	return true;
}

bool BinaryDataRead::rereadFindStepData( int find_id, int section )
{
	//int section = 0;
	//if(!skipStepHeader(section, 0)){
	//	return false;
	//}
	
	// STEP SIZE DATA
	ulongint_t udatasize = 0;
	if(!_binaryIO.readUINT64(udatasize,_binaryOpened)) return false;
	
	// NUMBER
	unsigned int counter = 0;
	if(!_binaryIO.readUINT32(counter,_binaryOpened)) return false;
	
	//////////////////////////////////////////////////////////////
	const char* name = 0;
	unsigned char typeID = 0;
	unsigned char partArrayDim = 0;
	int id;
	int parentID = 0;
	StructureCacheComponent* component;
	bool found_data = false;
	for(unsigned int i=0;i<counter;i++){
		// ID
		if(!_binaryIO.readINT32(id,_binaryOpened)) return false;
		if(id > 0){
			// ParentID
			if(!_nameManager->getBinaryAttribute(id,&name,&parentID,&typeID,&partArrayDim)){
				setErrorMessage( 10 ); // Collapsed.
				return false;
			}
		}
		// Component.
		component = _nameManager->findComponentFromFullName(name);
		if(id != find_id || component == 0){
			// ignore.
			unsigned short fullArrayDim = _nameManager->getFullArrayDim(id);
			if(!skipSection(typeID, partArrayDim, fullArrayDim)) return false;
			continue;
		}
		found_data = true;
		break;
	}
	if(!found_data){
		return false;
	}
	// section.
	return rereadSection(typeID,section,component);
}

bool BinaryDataRead::skipArraySize(unsigned char adim)
{
	// Data size
	if(adim > 0){
		unsigned int number = 0;
		// Array index data.
		unsigned int asize = 0;
		for(unsigned short i=0;i<adim;i++){
			if(!_binaryIO.readUINT32(asize,_binaryOpened)) return false;
			for(unsigned int j=0;j<asize;j++){
				if(!_binaryIO.readUINT32(number,_binaryOpened)) return false;
			}
		}
	}
	return true;
}

bool BinaryDataRead::rereadSection(
	unsigned char typeID,
	int section,
	StructureCacheComponent* component
)
{
	if(!component) return false;
	CacheSection* csection = component->section(section);
	if(!csection) return false;
	///////////////////////////////////////////////////////
	BaseDataType baseType = component->attribute()->baseType();
	int dataSize = 0;
	if(typeID > 0){
		const char* typeName = _binaryInfo->getTypeName(typeID);
		if(!_binaryInfo->dataTypeNameComp(baseType,typeName)) return false;
		dataSize = _binaryInfo->getTypeByteFromID(typeID);
	}

	// Array dimension
	unsigned char adim = (unsigned char)component->attribute()->arrayDim();

	// Data size
	skipArraySize(adim);
	
	// Values.
	return rereadValue(typeID,baseType,csection);
}

bool BinaryDataRead::rereadValue(
	unsigned char typeID,
	BaseDataType baseType,
	CacheSection* csection
)
{
	if(!entityType(baseType)) return true;
	if(csection->isArray()){
		unsigned int nsize = 0;
		// Number of data.
		if(!_binaryIO.readUINT32(nsize,_binaryOpened)) return false;
		LimitArraySection* arsection = (LimitArraySection*) csection;
		if (nsize == 0) {
			arsection->setSizeOfRead(0);
			return true;
		}
		unsigned int num_read = arsection->sizeOfRead();
		if(num_read == 0){
			return true;
		}
		unsigned int num_reread = nsize - num_read;
		// STRING.
		if(stringType(baseType)){
			// Skip to the reading position
			skipBinarySTRINGNL( num_read, _binaryOpened );
			// Reread
			char* value;
			bool rereading = arsection->isRereading();
			arsection->setRereading(true);
			for(unsigned int i=num_read;i<nsize;i++){
				if(readBinarySTRINGNL(&value,_binaryOpened) < 0){
					arsection->setRereading(rereading);
					return false;
				}
				if(!csection->setData(i,value,true)){
					arsection->setRereading(rereading);
					return false;
				}
			}
			arsection->setRereading(rereading);
		}
		// INTEGER.
		else if(integerType(baseType)){
			unsigned char nflag;
			switch(_binaryInfo->getTypeByteFromID(typeID)){
			case 1:
				{
					// Skip to the reading position
					skipArray(num_read, 1, _binaryOpened);
					// Reread
					BYTE_TYPE value;
					bool rereading = arsection->isRereading();
					arsection->setRereading(true);
					for(unsigned int i=num_read;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readUCHAR(value,_binaryOpened);
						if(!csection->setData(i,value,true)){
							arsection->setRereading(rereading);
							return false;
						}
					}
					arsection->setRereading(rereading);
				}
				break;
			case 2:
				{
					// Skip to the reading position
					skipArray(num_read, 2, _binaryOpened);
					// Reread
					SHORT_TYPE value;
					bool rereading = arsection->isRereading();
					arsection->setRereading(true);
					for(unsigned int i=num_read;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readINT16(value,_binaryOpened);
						if(!csection->setData(i,value,true)){
							arsection->setRereading(rereading);
							return false;
						}
					}
					arsection->setRereading(rereading);
				}
				break;
			case 4:
				{
					// Skip to the reading position
					skipArray(num_read, 4, _binaryOpened);
					// Reread
					INT_TYPE value;
					bool rereading = arsection->isRereading();
					arsection->setRereading(true);
					for(unsigned int i=num_read;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readINT32(value,_binaryOpened);
						if(!csection->setData(i,value,true)){
							arsection->setRereading(rereading);
							return false;
						}
					}
					arsection->setRereading(rereading);
				}
				break;
			default:
				return false;
			}
		}
		// FLOAT.
		else if(realType(baseType)){
			unsigned char nflag;
			switch(_binaryInfo->getTypeByteFromID(typeID)){
			case 4:
				{
					// Skip to the reading position
					skipArray(num_read, 4, _binaryOpened);
					// Reread
					float value;
					bool rereading = arsection->isRereading();
					arsection->setRereading(true);
					for(unsigned int i=num_read;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readREAL32(value,_binaryOpened);
						if(!csection->setData(i,value,true)){
							arsection->setRereading(rereading);
							return false;
						}
					}
					arsection->setRereading(rereading);
				}
				break;
			case 8:
				{
					// Skip to the reading position
					skipArray(num_read, 8, _binaryOpened);
					// Reread
					DOUBLE_TYPE value;
					bool rereading = arsection->isRereading();
					arsection->setRereading(true);
					for(unsigned int i=num_read;i<nsize;i++){
						if(_binaryInfo->existNullFlag()) _binaryIO.readUCHAR(nflag,_binaryOpened);
						_binaryIO.readREAL64(value,_binaryOpened);
						if(!csection->setData(i,value,true)){
							arsection->setRereading(rereading);
							return false;
						}
					}
					arsection->setRereading(rereading);
				}
				break;
			default:
				return false;
			}
		} // End of TYPE.

		/* Special operation for multi-array is needed.
		if(arsection->partArrayDimension() > 1){
			UtilArrayEquilibration equiliblium;
			if(equiliblium.setup(arsection)){
				equiliblium.operation();
			}
		}
		*/
		// Reset size of read.
		arsection->setSizeOfRead( 0 );
	} // End of FULL ARRAY.
	else{ // SCALAR.
		return false;
	} // END of SCALAR.
	return true;
}

} // END of namespace.

