
/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2007, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *  
 *  $Id: CacheRecordOperation.h,v 1.10 2007/07/30 03:29:23 nishimoto Exp $
 */

#ifndef CACHE_RECORD_OPERATION_H
#define CACHE_RECORD_OPERATION_H

#include <string.h>
#include <string>
#include <vector>
#include "util_sorted_vector.h"
#include "cache_file_utility.h"

#define CACHE_RECORD_OPERATION_OLD    0
#define CACHE_RECORD_OPERATION_NEW    1
#define CACHE_RECORD_OPERATION_ERASE  2
#define CACHE_RECORD_SHIFT_NOVALUE ((unsigned int)(-1))
#define CACHE_RECORD_NO_RECORD     ((unsigned int)(-1))

// function.
char* fcache_record_operation_strdup( const char* ss );
bool fcache_record_operation_file_exist( const char* filename );
void fcache_record_divide_filename( char* buffer, unsigned int record, int number );
bool fcache_make_output_directory( const char* outfile, const char* post_fix, std::string& outpath );
bool fcache_make_record_divide_filename( std::string& name, const std::string& directory, unsigned int record, bool check_exist );

// class.
class CacheEditedRecord {
private:
	unsigned int original_record_number;
	char* file_path;
	char* record_label;
public:
	CacheEditedRecord( unsigned int record );
	~CacheEditedRecord();
	// Clear.
	void clear();
	void deleteFile();
	// Set.
	void setRecordNo( const unsigned int record );
	void setFile( const char* path );
	void setLabel( const char* label );
	// Get.
	unsigned int getRecordNo() const {
		return original_record_number;
	}
	bool sameLabel( const char* label ){
		if(!label) return false;
		if(!record_label) return false;
		return ( strcmp(record_label,label) == 0 );
	}
	char* getLabel() const {
		return record_label;
	}
	char* getFile() const {
		return file_path;
	}
};

// Edited container.
class CacheCompareRecordNumber : public util_sorted::CompareClass<CacheEditedRecord> {
public:
	bool operator()(
		const CacheEditedRecord* a,
		const CacheEditedRecord* b
	)
	{
		return ( a->getRecordNo() < b->getRecordNo() );
	}
};
class CacheEqualLabel : public util_sorted::EqualClass<CacheEditedRecord> {
public:
	char* _label;
	CacheEqualLabel( const char* label ){
		_label = NULL;
		if(label){
			int len = strlen(label);
			char* pp = new char[len + 1];
			strcpy(pp,_label);
		}
	}
	virtual ~CacheEqualLabel(){
		if(_label) delete[] _label;
	}
	bool operator()( const CacheEditedRecord* a ) {
		if(!_label) return false;
		if(!a) return false;
		const char* lbl = a->getLabel();
		if(!lbl) return false;
		return (strcmp(lbl,_label) == 0);
	}
};

class CacheChangeRecord {
private:
	unsigned char change_type; // CACHE_RECORD_OPERATION_OLD, CACHE_RECORD_OPERATION_NEW.
	// unsigned int serial_record_number; -> serial_shift_record_number + array-index.
	union InsertedOrOriginalRecord {
		// Insert or Append record : change_type = [CACHE_RECORD_OPERATION_NEW]
		struct InsertedRecord {
			char* record_label; // For inserted record.
			char* file_path; // For inserted record & original edited record.
		}	insert;
		// Old but moved record : change_type = [CACHE_RECORD_OPERATION_OLD]
		struct OriginalRecord {
			//char* file_path; // For inserted record & original edited record.
			unsigned int original_record_number; // For original edited record.
		}	original;
	}	record;
public:
	CacheChangeRecord(unsigned char change_type);
	~CacheChangeRecord();
	void deleteFile();
	// Input.
	bool setOld(unsigned int original_record); //, char* file);
	bool setNew(const char* label, char* file);
	bool setLabel( const char* label );
	bool setFile( const char* label );
	//
	// Output.
	//
	unsigned char getType();
	bool isOld();
	bool isNew();
	bool sameLabel( const char* label );
	char* getLabel();
	char* getFile();
	unsigned int getOriginalRecordNumber();
};

class CacheRecordWork {
private:
	// original udf-file.
	unsigned int original_number_of_records;
	// Edit record.
	unsigned int minimum_edit_record; // Not original record;
	//unsigned int maximum_edit_record;
	//unsigned int maximum_erase_record;
	//
	// record operations.
	//
	unsigned int serial_shift_record_number; // Start changed or erased record.
	std::vector<CacheChangeRecord*> change_records;
	//
	// Edited registration.
	//
	util_sorted::util_sorted_vector<CacheEditedRecord,CacheCompareRecordNumber,CacheEqualLabel> *edited_records;
	//
	// methods.
	//
	void clearShiftRecordNumber();
	bool isShiftRecordNumber();
	void setShiftRecordNumber(unsigned int shift);
	unsigned int getShiftRecordNumber();
	void clearChangedRecord();
	void deleteRecord( CacheChangeRecord* rec );
	//-char* getFileNumber();
	void eraseEditedRecordByEraseRecordOperation( unsigned int original );
	// MIN-MAX RECORD.
	void check_minimum_edit_record(unsigned int record);
public:
	CacheRecordWork(unsigned int number_of_records);
	~CacheRecordWork();
	void deleteFiles();
	void clearRecordOperation();
	//
	// Input.
	//
	bool eraseRecord( unsigned int start, unsigned int number );
	bool insertRecord( unsigned int start, unsigned int number, bool front );
	bool appendRecord( unsigned int number );
	bool setRecordLabel( unsigned int record, const char* label );
	//
	// Output.
	//
	// Query.
	bool isChanged();
	bool isEdited();
	bool isRecordChanged();
	// Record operation (Insert/Append/Erase)
	unsigned int getRecordNumber();
	unsigned int getStartChangeRecord(); // FOR OUTPUT FILE.
	unsigned int getStartNewRecord();    // FOR OUTPUT FILE. AFTER THIS RECORD, THESE ARE ONLY APPENDED RECORDS.
	unsigned int getMaxOriginalChangedRecord(); // Return changed record index.

	// 0x0: No record (SKIP), 'O': OLD record, 'I': INSERTED record.
	unsigned char getStatus( unsigned int record );
	bool isOriginal( unsigned int record ){
		return ( 'O' == getStatus( record ) );
	}
	// OLD record
	unsigned int getOriginalRecord( unsigned int record );
	unsigned int getErasedOriginalRecords(std::vector<unsigned int>& records);

	// Edited old/Appended.
	char* getRecordLabel( unsigned int record ); // FOR NEW RECORD.
	// Edited old/Appended.
	bool findRecordFromLabel( const char* label, unsigned int& record );
	CacheChangeRecord* getChangeRecord( unsigned int record );
	void debugPrint();
	//
	// Edit flag.
	//
	// Record moved.
	bool isMovedRecord( unsigned int record );
	// Edited old/Appended.
	bool setEditRecord( unsigned int record, const char* label, const char* filepath = 0x0 );
	// Edited old/Appended.
	bool isEditedRecord( unsigned int record );
	// Edited old/Appended.
	char* getEditedRecordFilePath( unsigned int record );
	void setEditedRecordFilePath( unsigned int record, const char* filepath );
	// Only edited old record.
	CacheEditedRecord* getEditRecord( unsigned int record );

	//
	// Number
	//
	bool isAppendOnly();
	unsigned int sizeEditRecord();
	unsigned int sizeChangeRecord();	
};

//
// Changed records (Insert/Delete/Append) interface class.
//
class CacheRecordOperation {
private:
	unsigned int work_number_of_records;
	CacheRecordWork* _cache_record_work;

	// Common record.
	CacheEditedRecord* _common_record;
	bool is_common();
	void check_common();
	void delete_common();
	// Temporary directory.
	bool _set_record_temporary_directory;
	std::string _record_temporary_directory;

	void check_object();
	void delete_object();
public:
	CacheRecordOperation(unsigned int number_of_records);
	~CacheRecordOperation();
	void clearRecordOperation(bool clear_commmon = true);

	//
	// Temporary directory.
	//
	bool directory();
	void registDirectory();
	void clearDirectory();
	std::string &getDirectory();
	void setDirectory(std::string& dirpath);
	//
	// Input.
	//
	bool eraseRecord(unsigned int start, unsigned int number);
	bool insertRecord(unsigned int start, unsigned int number, bool front);
	bool appendRecord(unsigned int number);
	bool setRecordLabel(unsigned int record, const char* label);
	//
	// Output.
	//
	// Query.
	bool isChanged();
	bool isEdited();
	bool isRecordEdited(); // NO CHECK COMMON.
	bool isRecordChanged(); // NO CHECK COMMON.
	// Record operation (Insert/Append/Erase)
	unsigned int getRecordNumber();
	unsigned int getStartChangeRecord(); // FOR OUTPUT FILE.
	unsigned int getStartNewRecord(); // FOR OUTPUT FILE. AFTER THIS RECORD, THESE ARE ONLY APPENDED RECORDS.
	unsigned int getMaxOriginalChangedRecord(); // Return changed record index.

	// 0x0: No record (SKIP), 'O': OLD record, 'I': INSERTED record.
	//unsigned char getStatus( unsigned int record );
	bool isOriginal(unsigned int record);
	// OLD record
	unsigned int getOriginalRecord(unsigned int record);
	unsigned int getErasedOriginalRecords(std::vector<unsigned int>& records);
	// Get the original record number after editing records by a simple function.
	unsigned int getOriginalRecordSimpleWay(unsigned int record);
	// Edited old/Appended.
	char* getRecordLabel(unsigned int record); // FOR NEW RECORD.
	// Edited old/Appended.
	bool findRecordFromLabel(const char* label, unsigned int& record);
	//CacheChangeRecord* getChangeRecord( unsigned int record );
	void debugPrint();
	//
	// Edit flag.
	//
	// Edited old/Appended.
	bool setEditRecord(unsigned int record, const char* label, const char* filepath = 0x0);
	// Record moved. Return false for record = 0, true for INSERTED or MOVED ORIGINAL. 
	bool isMovedRecord(unsigned int record);
	// Edited old/Appended.
	bool isEditedRecord(unsigned int record);
	// Edited old/Appended.
	char* getEditedRecordFilePath(unsigned int record);
	void setEditedRecordFilePath(unsigned int record, const char* filepath);
	// Only edited old record.
	//CacheEditedRecord* getEditRecord( unsigned int record );
	//
	// Number
	//
	void resetCommonEdited();
	bool isCommonEdited();
	bool isAppendOnly();
	unsigned int sizeEditRecord();
	unsigned int sizeChangeRecord();
};

//
// For divided records operations.
//
class REC_LABELPOS_DATA;

class CacheDivideRecordData {
private:
	bool move_or_copy_file; // move : true , copy : false.
	bool executed_move_or_copy_or_delete;
	unsigned int record_number;
	char* source_file_path; // source file path.
	char* record_file_name; // destination file name, not full-path.
	char* record_label;
public:
	CacheDivideRecordData( unsigned int record );
	~CacheDivideRecordData();
	// Move or Copy file.
	void setMoveFile( bool is_move );
	bool isMoveFile();
	bool isExecuted();
	void setExecuted( bool executed );
	// Record number, >= 1.
	unsigned int getRecordNo() const;
	// Source.
	void setSourceFilePath( const char* file );
	const char* getSourceFilePath();
	// Destination.
	void setFileName( const char* file );
	const char* getFileName();
	// Label.
	void setLabel( const char* label );
	const char* getLabel();
};

class CacheDividedRecordOperation {
	//-- Inner class.
	class CompareRecordNumber : public util_sorted::CompareClass<CacheDivideRecordData> {
	public:
		bool operator()(
			const CacheDivideRecordData* a,
			const CacheDivideRecordData* b
		)
		{
			return ( a->getRecordNo() < b->getRecordNo() );
		}
	};
	class EqualLabel : public util_sorted::EqualClass<CacheDivideRecordData> {
	public:
		char* _label;
		EqualLabel( const char* label ){}
		virtual ~EqualLabel(){}
		bool operator()( const CacheDivideRecordData* a ) {
			return true;
		}
	};
private:
	typedef util_sorted::util_sorted_vector<CacheDivideRecordData,CompareRecordNumber,EqualLabel> msorted_vector;
	// Instance variables.
	//unsigned int num_record;
	//std::string original_filepath;
	bool overwrite;
	unsigned int link_file_with_directory_record; // "directory/rec00000001.du"
	std::string source_filepath;
	std::string output_filepath;
	std::string output_directory_path; // May be deleted. (all records are deleted.)
	std::string output_directory_name; // Only name, not path.
	msorted_vector* divide_records;
	std::vector<char*> delete_files; // file-path.

	// Objects in FileStreamManager.
	CacheRecordOperation* record_opp; // DO NOT FREE. (REFERRED)
	REC_LABELPOS_DATA*    record_dtp; // DO NOT FREE. (REFERRED)

	// methods.
	bool makeSourceFilePath(
		const char* record_file,
		std::string& file_path
	);
	void addDivideFile(
		unsigned int rec,
		const char* source_file,
		const char* destin_file,
		const char* label,
		bool is_move
	);
	void addDeleteFilePath(
		const char* filepath
	);
	bool executeOperation( CacheDivideRecordData* drecp );
public:
	CacheDividedRecordOperation(
		CacheRecordOperation* recop,
		REC_LABELPOS_DATA*    recdt
	);
	~CacheDividedRecordOperation();
	void makeDivideRecordData();
	// file.
	void setSourceOutputFile(
		bool overwrite,
		const char* source_filepath,
		const char* output_filepath
	);
	// directory.
	void setOutputDirectoryPath( const char* directory );

	// Output path / directory-name + file.
	// directory-PATH + delimiter + filename.
	bool getFilePath(
		std::string& outpath,
		const char* filename,
		const char* delimiter = NULL // NULL : depended on OS.
	);
	// directory-NAME + delimiter + filename
	bool getDirectoryFile(
		std::string& directory_file,
		const char* filename,
		const char* delimiter = NULL // NULL : depended on OS.
	);
	//--
	bool isLinkFileWithDirectory();
	unsigned int getStartLinkFileWithDirectory();
	void setRecordLinkFileWithDirectory( unsigned int record );

	// start devided record.
	unsigned int startDevidedRecord();
	// new divided record files.
	unsigned int sizeDividedRecord();
	CacheDivideRecordData* getDivideRecord( unsigned int record );
	// Delete files.
	unsigned int sizeDeleteFile();
	bool deleteFiles();
	// Execute.
	bool executeFileOperation( unsigned int record );
	bool executeRemainingFileOperation();
};

#endif // End of CACHE_RECORD_OPERATION_H

