/*
 * LimitArraySection
 */

#ifndef LIMIT_ARRAY_SECTION_H
#define LIMIT_ARRAY_SECTION_H
#include "UtilCoreCacheValue.h"
#include "CacheEnviron.h"
#include "RereadDataListener.h"

namespace CacheUtilCore {

#define MAX_LIMIT_SIZE	UINT_MAX

/**
 *  Reread listener to be called when array data not yet read is needed
 */
class RereadLimitArrayListener {
	int id;
	RereadDataListener* listener;
public:
	RereadLimitArrayListener(int id_):id(id_),listener(0){}
	RereadLimitArrayListener(int id_, RereadDataListener* listener_):id(id_),listener(listener_){}
	~RereadLimitArrayListener(){}
	void setListener(RereadDataListener* listener_){
		listener = listener_;
	}
	/**
	 * @brief Reread array data
	 * @param[in] (section) 0:common, 1:record
	 * @return true for normal end
	 */
	bool reread(unsigned int section){
		if(listener){
			return listener->reread(id, section);
		}
		return false;
	}
};

/**
 *  ArraySection with limited reading array data function
 */
class LimitArraySection : public ArraySection {
	bool is_rereading;
	unsigned int size_of_read_array;
	RereadLimitArrayListener* reread_listener;
	
public:
	LimitArraySection():
	is_rereading(false),
	size_of_read_array(0),
	reread_listener(0){}
	virtual ~LimitArraySection(){}

	void resetLimitArraySection() {
		size_of_read_array = 0;
	}
	virtual void resetArraySection() {
		ArraySection::resetArraySection();
		resetLimitArraySection();
	}
	/**
	 * @brief Set the limit size of reading array data
	 * @param[in] (size_of_read) size to read
	 */
	void setSizeOfRead(unsigned int size_of_read){
		size_of_read_array = size_of_read;
	}
	/**
	 * @brief Get the limit size of reading array data
	 * @return size to read
	 */
	unsigned int sizeOfRead(){
		return size_of_read_array;
	}
	/**
	 * @brief Get real size of reading array data
 	 * @param[in] (nsize) array size
 	 * @param[in] (is_partial_read) whether limited partial reading is valid
 	 * @param[in] (limit_size) limit size of reading
	 * @return real size to read
	 */
	unsigned int getRealSizeOfRead(unsigned int nsize, bool is_partial_read, unsigned int limit_size){
		if(!is_partial_read){
			setSizeOfRead(0);
			return nsize;
		}
		if((nsize / 2U) <= limit_size){
			setSizeOfRead(0);
			return nsize;
		}
		else if (isMultiArray()) {
			setSizeOfRead(0);
			return nsize;
		}
		else {
			setSizeOfRead(limit_size);
			return limit_size;
		}
	}
	/**
	 * @brief Set reread listener
 	 * @param[in] (reread_listener_) reread listener
	 */
	void setRereadListener(int id, RereadDataListener* listener);
	/**
	 * @brief Fire reread event
	 */
	bool reread();
	/**
	 * @brief whether is in rereading
	 * @return true for rereading
	 */
	bool isRereading(){
		return is_rereading;
	}
	/**
	 * @brief Set rereading switch
 	 * @param[in] (status) true means in rereading
	 */
	void setRereading(bool status){
		is_rereading = status;
	}

private:
	void setRereadListener(RereadLimitArrayListener* reread_listener_){
		reread_listener = reread_listener_;
	}
};

} // end of namespace.

#endif
