
/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *
 *  The structured multi-dimensional array data model.
 *  
 *  $Id: StructureCacheOperation.cpp,v 1.7 2007/05/25 13:36:17 nishimoto Exp $
 */

#include "StructureCacheOperation.h"
#include "StructureCacheComponent.h"
#include "StructureCacheGroup.h"
#include "StructureCacheData.h"
#include "BinaryDataManager.h"
#include "RereadDataListener.h"
#include "UtilArrayEquilibration.h"
using namespace CacheUtilCore;

// StructureCacheOperation
StructureCacheOperation::StructureCacheOperation(
	const StructureCacheOption& option,
	RereadDataListener* listener
)
{
	_cache_owner = true;
	_flag_final_parse = true;
	_section_id = 0;
	_cache = new StructureCacheData(option, listener);
	_tableIndex = new CacheUtilCore::CacheTableData();
	_binary_mgr = new BinaryDataManager(_cache);
}

StructureCacheOperation::StructureCacheOperation(StructureCacheData* cache)
{
	_cache_owner = false;
	_flag_final_parse = true;
	_section_id = cache->status()->current();
	_cache = cache;
	_tableIndex = new CacheTableData();
	_binary_mgr = new BinaryDataManager(_cache);
}

StructureCacheOperation::~StructureCacheOperation()
{
	if(_cache_owner) ownerDestroy();
	if(_tableIndex) delete _tableIndex;
	if(_binary_mgr) delete _binary_mgr;
}

// BinaryI/O [inputKey:"1:abc:123"]
void StructureCacheOperation::setInputKey( const char* inputKey )
{
	if(!inputKey) return;
	if(_binary_mgr){
		_binary_mgr->setInputKey( inputKey );
	}
}
void StructureCacheOperation::pushOpenKeys( map<string,int>* _keys )
{
	if(!_keys) return;
	if(_binary_mgr){
		_binary_mgr->pushOpenKeys( _keys );
	}
}
const char* StructureCacheOperation::getKeyWord()
{
	if(!_binary_mgr) return 0;
	return _binary_mgr->getKeyWord();
}
int StructureCacheOperation::getKeyCode()
{
	if(!_binary_mgr) return 0;
	return _binary_mgr->getKeyCode();
}

void StructureCacheOperation::ownerDestroy()
{
	delete _cache;
}
StructureCacheData* StructureCacheOperation::getCacheData()
{
	return _cache;
}

void
StructureCacheOperation::clearStructure()
{
	_cache->clearStructure();
	resetTable();
	_binary_mgr->clearBinaryDataManager();
}

void
StructureCacheOperation::clearData()
{
	_cache->clearData(section());
}
void
StructureCacheOperation::clearData( int section )
{
	if(section < 0) _cache->clearData(-1);
	else            _cache->clearData(section);
	resetTable();
}

void 
StructureCacheOperation::changeEvent()
{
	resetTable();
}

int StructureCacheOperation::dimSection()
{
	if(!_cache) return 0;
	StructureCacheStatus* stst = _cache->status();
	if(stst) return stst->number();
	return 0;
}

///////////// Define ///////////////
bool StructureCacheOperation::insertDefine(
	int          id,
	int          parentID,
	const char*  fullName,
	BaseDataType etype,
	unsigned short partDimension
)
{
	return _cache->insertDefine(id,parentID,fullName,etype,partDimension);
}

bool StructureCacheOperation::insertDefine(
	int          id,
	int          parentID,
	const char*  fullName,
	BaseDataType etype
)
{
	unsigned short partDimension = TypeInfo::getLastArrayDimention(fullName);
	return _cache->insertDefine(id,parentID,fullName,etype,partDimension);
}

bool StructureCacheOperation::setInitialFixArraySize(
	int          tergetID,
	unsigned int index,
	unsigned int number,
	int          setID,
	const int    section
)
{
	StructureCacheComponent* component = _cache->getComponent(tergetID);
	if(!component) return false;
	StructureCacheComponent* setComponent = 0;
	if(setID > 0){
		setComponent = _cache->getComponent(setID);
	}
	return SectionTool::setInitialFixArraySize(component,index,number,setComponent,section);
}

// Process of parsing.
void StructureCacheOperation::setStatus(
	const unsigned int sectionNumber,
	const GlobalArraySizeSpecType spec
)
{
	StructureCacheStatus* status =_cache->status();
	if(sectionNumber > 0){
		status->changeNumber(sectionNumber);
	}
	if(spec != globalArrayIgnoreSize){
		status->changeArraySizeSpec(spec);
	}
}

void StructureCacheOperation::finalParse()
{
	if(!_flag_final_parse) return;
	StructureCacheStatus* status =_cache->status();
	switch(status->getArraySizeSpec()){
	case globalArrayRigidSize:
	case globalArrayMultiRigid:
		//UtilCoreCacheArrayTool::iterateSetupMaxSize(_cache->getComponent(),this->section());
		UtilCoreCacheArrayTool::iterateEquilibration(_cache->getComponent(),this->section());
		break;
	}
}

////////////// Size ////////////////
// Section.
bool StructureCacheOperation::hasData()
{
	return hasData(this->section());
}
bool StructureCacheOperation::hasData(
	int section
)
{
	StructureCacheComponent* component = _cache->getComponent(0); // Virtual Top component.
	//if(component->sequenceSize(section) > 0) return true; return false;
	return component->existData(section,0);
}
// Component.
bool StructureCacheOperation::existData(
	int id,
	const UtilIndexVectorType* index, // = 0,
	SectionPriorityType priority,     // = PRIORITY_RECORD_EXT,
	unsigned int*       gotSection    // = 0x0
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	bool rb = false;
	for(sectionIt.First(section(),priority);sectionIt.IsDone();sectionIt.Next()){
		if(sectionIt.Count() > 0){
			if(!(rb = component->existData(sectionIt.CurrentItem(),index))) continue;
		}
		if(gotSection){
			*gotSection = sectionIt.CurrentItem();
		}
		return rb;
	}
	return rb;
}


unsigned int StructureCacheOperation::size(
	int id,
	SectionPriorityType priority,
	unsigned int*       gotSection
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return 0;
	unsigned int nsize = 0;
	for(sectionIt.First(section(),priority);sectionIt.IsDone();sectionIt.Next()){
		if(sectionIt.Count() > 0){
			if((nsize = component->sequenceSize(sectionIt.CurrentItem())) <= 0) continue;
		}
		if(gotSection){
			*gotSection = sectionIt.CurrentItem();
		}
		return nsize;
	}
	return 0;
}

bool StructureCacheOperation::getSize(
	const int id,
	const UtilIndexVectorType&	data_index,	// IN: INDEX
	unsigned int&		nsize,		// OUT: SIZE OF ELEMENTS
	SectionPriorityType	priority,
	unsigned int*       gotSection,
	unsigned short      reduceIndexDim,
	unsigned short      reduceArrayDim,
	const bool          useExtension
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	for(sectionIt.First(section(),priority);sectionIt.IsDone();sectionIt.Next()){
		if(sectionIt.Count() > 1){
			//if(component->sequenceSize(sectionIt.CurrentItem()) <= 0) continue;
			if(!component->existData(sectionIt.CurrentItem(),&data_index)) continue;
		}
		if(component->indexSize(sectionIt.CurrentItem(),data_index,nsize,reduceIndexDim,reduceArrayDim,useExtension)){
			if(gotSection){
				*gotSection = sectionIt.CurrentItem();
			}
			return true;
		}
	}
	nsize = 0;
	return true;
}
bool StructureCacheOperation::getSize( 
	const int id,
	const UtilIndexVectorType& data_index,	// IN: INDEX
	unsigned int&              nsize,		// OUT: SIZE OF ELEMENTS
	int                        section,
	unsigned short             reduceIndexDim,
	unsigned short             reduceArrayDim,
	const bool                 useExtension
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	if(section < 0) section = this->section();
	return component->indexSize(section,data_index,nsize,reduceIndexDim,reduceArrayDim,useExtension);
}

bool StructureCacheOperation::getSizeList(
	const int      id,               // IN
	UtilIndexVectorType& data_size,  // OUT
	int            section,              // IN
	const bool     forceToGet
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	if(section < 0) section = this->section();
	CacheSection* csect = component->section(section);
	if(!csect) return false;
	return SectionTool::getArraySizeList(*csect,data_size,forceToGet);
}

bool StructureCacheOperation::getSizeList(
	const int      id,         // IN
	UtilIndexVectorType& data_size, // OUT
	SectionPriorityType  priority,  // IN
	const bool           forceToGet
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	for(sectionIt.First(section(),priority);sectionIt.IsDone();sectionIt.Next()){
		if(sectionIt.Count() > 1){
			//if(component->sequenceSize(sectionIt.CurrentItem()) <= 0) continue;
			if(!component->existData(sectionIt.CurrentItem(),0)) continue;
		}
		CacheSection* csect = component->section(sectionIt.CurrentItem());
		if(!csect) return false;
		if(SectionTool::getArraySizeList(*csect,data_size,forceToGet)){
			return true;
		}
	}
	data_size.resize(component->attribute()->fullDim(),0);
	return true;
}

// Make an uneven multi-dimensional array to an equilibrium one.
bool StructureCacheOperation::equilibrateMatrix(
	const int      id,        // IN
	const int      section,   // IN
	const UtilIndexVectorType* index // = 0
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	CacheSection* csect = 0;
	if(section < 0) csect = component->section(this->section());
	else            csect = component->section(section);
	if(!csect) return false;
	UtilArrayEquilibration equiliblium;
	if(equiliblium.setup(csect,index)){
		return equiliblium.operation();
	}
	return false;
}

// Row size.
bool StructureCacheOperation::getRowSize(
	const int id,
	const UtilIndexVectorType& data_index,	// IN: INDEX
	unsigned int&       nsize,		// OUT: SIZE OF ELEMENTS
	int section
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	if(section < 0) section = this->section();
	CacheSection* sect = component->section(section);
	if(!sect) return false;
	return SectionTool::getRowSizeIndex(*sect,data_index,nsize);
}

////////////// Insert/Delete //////////////
bool StructureCacheOperation::insertData(
	const int section,
	const int id,
	const UtilIndexVectorType&	data_index,	// IN: INDEX
	const unsigned int number,				// IN: number to insert.
	const bool force,
	const bool repeat
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->insertData(section,data_index,number,true,force,repeat);
}
bool StructureCacheOperation::insertData(
	const int id,
	const UtilIndexVectorType&	data_index,	// IN: INDEX
	const unsigned int number,				// IN: number to insert.
	const bool force,
	const bool repeat
)
{
	return insertData(section(),id,data_index,number,force,repeat);
}
bool StructureCacheOperation::deleteData(
	const int section,
	const int id,
	const UtilIndexVectorType&	data_index,	// IN: INDEX
	const unsigned int		number		// IN: delete number	
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->deleteData(section,data_index,number);
}
bool StructureCacheOperation::deleteData(
	const int id,
	const UtilIndexVectorType&	data_index,	// IN: INDEX
	const unsigned int		number		// IN: delete number	
)
{
	return deleteData(section(),id,data_index,number);
}

///////////////////////////////////////
////////////// Set data. //////////////
///////////////////////////////////////
//////// With section number. /////////
bool StructureCacheOperation::setData(
	const int section,
	const int id,
	const UtilIndexVectorType& index,
	const char* data,
	const bool append
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->setData(section,index,data,append);
}
bool StructureCacheOperation::setData(
	const int section,
	const int id,
	const UtilIndexVectorType& index,
	const LONG_TYPE data,
	const bool append
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->setData(section,index,data,append);
}
bool StructureCacheOperation::setData(
	const int section,
	const int id,
	const UtilIndexVectorType& index,
	const DOUBLE_TYPE data,
	const bool append
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->setData(section,index,data,append);
}
bool StructureCacheOperation::setNull(
	const int section,
	const int id,
	const UtilIndexVectorType& index,
	const bool append
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->setNull(section,index,append);
}

///////////////////////////////////////
bool StructureCacheOperation::setData(
	const int id,
	const UtilIndexVectorType& index,
	const char* data,
	const bool append
)
{
	return setData(section(),id,index,data,append);
}
bool StructureCacheOperation::setData(
	const int id,
	const UtilIndexVectorType& index,
	const LONG_TYPE data,
	const bool append
)
{
	return setData(section(),id,index,data,append);
}
bool StructureCacheOperation::setData(
	const int id,
	const UtilIndexVectorType& index,
	const DOUBLE_TYPE data,
	const bool append
)
{
	return setData(section(),id,index,data,append);
}
bool StructureCacheOperation::setNull(
	const int id,
	const UtilIndexVectorType& index,
	const bool append
)
{
	return setNull(section(),id,index,append);
}

//////////////////////////////////////////
////////////// Append data. //////////////
//////////////////////////////////////////
//////// With section number. ///////////
bool StructureCacheOperation::appendData(
	const int section,
	const int id,
	const UtilIndexVectorType& index,
	const char* data,
	const bool append
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->appendData(section,index,data,append);
}
bool StructureCacheOperation::appendData(
	const int section,
	const int id,
	const UtilIndexVectorType& index,
	const LONG_TYPE data,
	const bool append
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->appendData(section,index,data,append);
}
bool StructureCacheOperation::appendData(
	const int section,
	const int id,
	const UtilIndexVectorType& index,
	const DOUBLE_TYPE data,
	const bool append
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->appendData(section,index,data,append);
}

// CHAR -> EACH TYPE.
bool StructureCacheOperation::appendConvertData(
	const int section,
	const int id,
	const UtilIndexVectorType& index,
	const char* data,
	const bool append
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->appendConvertData(section,index,data,append);
}

///////////////////////////////////////
bool StructureCacheOperation::appendData(
	const int id,
	const UtilIndexVectorType& index,
	const char* data,
	const bool append
)
{
	return appendData(section(),id,index,data,append);
}
bool StructureCacheOperation::appendData(
	const int id,
	const UtilIndexVectorType& index,
	const LONG_TYPE data,
	const bool append
)
{
	return appendData(section(),id,index,data,append);
}
bool StructureCacheOperation::appendData(
	const int id,
	const UtilIndexVectorType& index,
	const DOUBLE_TYPE data,
	const bool append
)
{
	return appendData(section(),id,index,data,append);
}
// CHAR -> EACH TYPE.
bool StructureCacheOperation::appendConvertData(
	const int id,
	const UtilIndexVectorType& index,
	const char* data,
	const bool append
)
{
	return appendConvertData(section(),id,index,data,append);
}

///////////// Get data. ///////////////
bool StructureCacheOperation::getData(
	const int id,
	unsigned int seq,
	CoreCacheDataInfo&	val
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	return component->getData(section(),seq,val);
}
bool StructureCacheOperation::getData(
	const int id,
	const UtilIndexVectorType& index,    // IN: INDEX
	CoreCacheDataInfo&         value,    // OUT:
	SectionPriorityType        priority, // IN: option
	unsigned int*              gotSection
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	value.fill = false;
	for(sectionIt.First(section(),priority);sectionIt.IsDone();sectionIt.Next()){
		if(sectionIt.Count() > 1){
			//if(component->sequenceSize(sectionIt.CurrentItem()) <= 0) continue;
			if(!component->existData(sectionIt.CurrentItem(),&index)) continue;
		}
		if(component->getData(sectionIt.CurrentItem(),index,value)){
			if(gotSection){
				if(value.fill) *gotSection = sectionIt.CurrentItem();
			}
			return true;
		}
	}
	return false;
}
bool StructureCacheOperation::getData(
	const int id,
	const UtilIndexVectorType& index,    // IN: INDEX
	CoreCacheUnionType&        value,    // OUT:
	bool&                      exist,
	SectionPriorityType        priority, // IN: option
	unsigned int*              gotSection
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	exist = false;
	for(sectionIt.First(section(),priority);sectionIt.IsDone();sectionIt.Next()){
		if(sectionIt.Count() > 1){
			//if(component->sequenceSize(sectionIt.CurrentItem()) <= 0) continue;
			if(!component->existData(sectionIt.CurrentItem(),&index)) continue;
		}
		if(component->getData(sectionIt.CurrentItem(),index,value,exist)){
			if(gotSection){
				if(exist) *gotSection = sectionIt.CurrentItem();
			}
			return true;
		}
	}
	return false;
}

//// Index-sequence relation.
bool 
StructureCacheOperation::indexToSequence(
	const int id,
	const UtilIndexVectorType& index,
	unsigned int& sequence,
	int section
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	if(section < 0) section = this->section();
	return component->indexToSequence(section,index,sequence);
}

bool 
StructureCacheOperation::sequenceToIndex(
	const int id,
	const unsigned int& sequence,
	UtilIndexVectorType& index,
	int section
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	if(section < 0) section = this->section();
	return component->sequenceToIndex(section,sequence,index);
}

// Table format index & size.
bool StructureCacheOperation::getTableSize(
	const int id,
	int section,
	const UtilIndexVectorType& inIndex,
	unsigned int& nsize
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	if(section < 0) section = this->section();
	CacheSection* sect = component->section(section);
	if(!sect) return false;
	//////////////////////////////////////////////////
	return _tableIndex->getSize(sect,inIndex,nsize);
}

bool 
StructureCacheOperation::getTableIndex(
	const int id,
	int section,
	const UtilIndexVectorType& inIndex,
	unsigned int               sequence,
	UtilIndexVectorType&       outIndex
)
{
	StructureCacheComponent* component = _cache->getComponent(id);
	if(!component) return false;
	if(section < 0) section = this->section();
	CacheSection* sect = component->section(section);
	if(!sect) return false;
	//////////////////////////////////////////////////
	return _tableIndex->getIndex(sect,inIndex,sequence,outIndex);
}

void StructureCacheOperation::resetTable()
{
	_tableIndex->clear();
}

void StructureCacheOperation::printTable()
{
	_tableIndex->printTable();
}
///////////////////////////////////////
//// Binary I/O ///////////////////////
///////////////////////////////////////
// Binary Information.
void StructureCacheOperation::setBigEndian()
{
	_binary_mgr->readyWrite();
	_binary_mgr->setBigEndian();
}
void StructureCacheOperation::setLittleEndian()
{
	_binary_mgr->readyWrite();
	_binary_mgr->setLittleEndian();
}

// WRITE.
bool StructureCacheOperation::writeBinaryHeader(
	FILE* binaryOpened
)
{
	_binary_mgr->readyWrite();
	return _binary_mgr->binaryWriteHeader(binaryOpened);
}
bool StructureCacheOperation::writeDivideHeader(
	FILE* binaryOpened,
	UserBinaryUdfData* user_data
)
{
	_binary_mgr->readyWrite();
	return _binary_mgr->binaryWriteDivideHeader(binaryOpened, user_data);
}
bool StructureCacheOperation::writeBinaryData(
	int section,
	const char* label,
	FILE* binaryOpened,
	bool emptyOutput
)
{
	_binary_mgr->readyWrite();
	if(emptyOutput){
		return _binary_mgr->binaryWriteEmpty(section,label,binaryOpened);
	}
	else{
		return _binary_mgr->binaryWriteData(section,label,binaryOpened);
	}
}

// READ.
bool StructureCacheOperation::readBinaryHeader(
	unsigned int* datasize,
	int*  shiftFineTuning,
	FILE* binaryOpened
)
{
	_binary_mgr->readyRead();
	unsigned short hsize = 0;
	if(_binary_mgr->binaryReadHeader(0,shiftFineTuning,binaryOpened)){
		if(datasize){
			*datasize = (unsigned int)hsize;
			// size of data types -> fullHeaderSize.
			_binary_mgr->fullHeaderSize(*datasize);
		}
		return true;
	}
	return false;
}

int StructureCacheOperation::readDivideKey(
	int*  shiftFineTuning,
	FILE* binaryOpened
)
{
	_binary_mgr->readyRead();
	return _binary_mgr->readDivideKey(shiftFineTuning,binaryOpened);
}
/*
void StructureCacheOperation::setIsUserDirectory(bool user)
{
	_binary_mgr->setIsUserDirectory(user);
}
void StructureCacheOperation::setIsUserRecord(bool user)
{
	_binary_mgr->setIsUserRecord(user);
}
bool StructureCacheOperation::isUserDirectory()
{
	return _binary_mgr->isUserDirectory();
}
bool StructureCacheOperation::isUserRecord()
{
	return _binary_mgr->isUserDirectory();
}
*/
void StructureCacheOperation::setDivideDirectory(const char* directory)
{
	_binary_mgr->setDivideDirectory(directory);
}
const char* StructureCacheOperation::getDivideDirectory()
{
	return _binary_mgr->getDivideDirectory();
}
void StructureCacheOperation::setDivideRecordPrefix(const char* record)
{
	_binary_mgr->setDivideRecordPrefix(record);
}
const char* StructureCacheOperation::getDivideRecordPrefix()
{
	return _binary_mgr->getDivideRecordPrefix();
}
const char* StructureCacheOperation::getDivideDrawObjectFile()
{
	return _binary_mgr->getDivideDrawObjectFile();
}

bool StructureCacheOperation::readBinaryData(
	int&       section,
	string&    label,
	longint_t& datasize,
	int*       shiftFineTuning,
	SortedDataPositionList* data_position_list,
	FILE*      binaryOpened
)
{
	_binary_mgr->readyRead();
	return _binary_mgr->binaryReadData(section,label,datasize,shiftFineTuning,data_position_list,binaryOpened);
}

bool StructureCacheOperation::rereadBinaryData(
	int id,
	SortedDataPositionList* data_position_list,
	FILE*      binaryOpened
)
{
	_binary_mgr->readyRead();
	return _binary_mgr->binaryRereadData(id,data_position_list,binaryOpened);
}

bool StructureCacheOperation::rereadBinaryAllData(
	FILE*      binaryOpened
)
{
	_binary_mgr->readyRead();
	return _binary_mgr->binaryRereadAllData(binaryOpened);
}

bool StructureCacheOperation::binaryCheckNeedReread(unsigned int section)
{
	if(_binary_mgr){
		return _binary_mgr->binaryCheckNeedReread(section);
	}
	return false;
}

bool StructureCacheOperation::rereadCheckAndAll(
	unsigned int section,
	unsigned int currentRecord,
	FileStreamManager* fsmgr
)
{
	if(binaryCheckNeedReread(section)){
		unsigned int cur_rec = currentRecord;
		// Get original record.
		CacheRecordOperation* recop = fsmgr->getRecordOperation();
		if (recop) {
			cur_rec = recop->getOriginalRecordSimpleWay(cur_rec);
		}

		bool is_virtual = fsmgr->isOpenedVirtual();
		string reread_file;
		const char* open_file = fsmgr->getOpenedFile();
		if (open_file) {
			reread_file.assign(open_file);
		}
		else { /* Normally, this process is not entered. */
			if (!fsmgr->getReadFile(cur_rec, reread_file, is_virtual)) {
				return false;
			}
		}

		BinaryInput binRead(is_virtual); // Close in destructor.
		if(!binRead.open(reread_file.c_str(), is_virtual)){
			return false;
		}
		return binRead.rereadAllData(cur_rec, fsmgr, this);
	}
	return false;
}

bool StructureCacheOperation::skipBinaryData(
	FileStreamManager* fmgr,
	int&       section,
	string&    label,
	longint_t& datasize,
	int*       shiftFineTuning,
	FILE*      binaryOpened
)
{
	_binary_mgr->readyRead();
	return _binary_mgr->binarySkipData(fmgr,section,label,datasize,shiftFineTuning,binaryOpened);
}
bool StructureCacheOperation::readBinarySeries(
	int&       section,
	string&    label,
	longint_t& datasize,
	FILE*      binaryOpened
)
{
	_binary_mgr->readyRead();
	return _binary_mgr->binaryReadSeries(section,label,datasize,binaryOpened);
}
const char* StructureCacheOperation::getBinaryReadError() const
{
	const char* msg = _binary_mgr->getReadError();
	if(!msg) return 0;
	if(*msg) return msg;
	return 0;
}

// Copy.
bool StructureCacheOperation::binaryCopy(
	FILE* destination,
	FILE* source,
	int*   section,
	string*    label,
	longint_t* datasize
)
{
	_binary_mgr->readyWrite(); // _binary is a BinaryDataManager instance.
	return _binary_mgr->binaryCopy(destination,source,section,label,datasize);
}
// The copy specified size or plain copy.
bool StructureCacheOperation::binaryCopySpec(
	FILE* destination,
	FILE* source,
	longint_t  copySize,
	longint_t* copiedSize
)
{
	return _binary_mgr->binaryCopySpec(destination,source,copySize,copiedSize);
}
