
/*
 *  Material Modeling Platform Package
 *  Copyright(c) 2002, The Japan Research Institute, Ltd.
 *  All rights reserved.
 *  
 *  $Id: binaryInput.cpp,v 1.4 2007/06/04 11:04:34 nishimoto Exp $
 */

#ifdef WIN32
#pragma warning (disable:4786)
#endif

#include "binaryConvert.h"
#include "binaryInput.h"
#include "textudfOutput.h"
#include "binaryOutput.h"
#include "FileStreamManager.h"
#include "MasterCacheMaker.h"
#include "ConvertBinaryReadUDFRecord.h"
#include "StructureCacheOperation.h"
#include "cache_util_character.h"
#include "BinaryFileLib.h"
#include "CacheExportManager.h"
#include "cache_directory_utility.h"
#include "BinaryDataRead.h"
#include "BinaryDataWrite.h"
using namespace std;
using namespace binaryio;

BinaryInput::~BinaryInput()
{
	close();
	destroyTextOutput();
	if(divide_record_data){
		delete divide_record_data;
	}
}

bool BinaryInput::isBinaryCommon()
{
	return existBinaryCommon;
}

// File operation.
FILE* BinaryInput::open( const char* file, bool open_virtual)
{
	if(!file) return 0;
#if defined(WIN32) || defined(WIN64) || defined(CYGWIN_WIN32PATH)
	_binaryOpend = fopen(file,"rb");
#else
	_binaryOpend = fopen(file,"r");
#endif
	if (_binaryOpend) {
		_binaryOpendFile.assign(file);
		_binaryOpendIsVirtual = open_virtual;
	}
	return _binaryOpend;
}


void BinaryInput::close()
{
	if(_binaryOpend) fclose(_binaryOpend);
	_binaryOpend = 0;
	_binaryOpendFile.clear();
}


bool BinaryInput::seek( longint_t pos )
{
	if(!_binaryOpend) return false;
	return UtilHugeFile::seek( _binaryOpend, pos, SEEK_SET ) == 0;
}

bool BinaryInput::seek( FILE* fp, longint_t pos )
{
	return UtilHugeFile::seek( fp, pos, SEEK_SET ) == 0;
}

longint_t BinaryInput::seekEnd()
{
	if(!_binaryOpend) return 0;
	if(UtilHugeFile::seek( _binaryOpend, 0L, SEEK_END ) != 0){
		return 0;
	}
	return ftell(_binaryOpend);
}

longint_t BinaryInput::tell()
{
	if(!_binaryOpend) return 0;
	return ftell(_binaryOpend);
}

bool BinaryInput::eof()
{
	return (feof(_binaryOpend) != 0);
}

// Optional operation.
void BinaryInput::setRecorder( AbstractReadUDFRecord* recorder )
{
	_read_recorder = recorder;
}

// Text output.
bool BinaryInput::createTextOutput( const char* file, bool appendMode )
{
	if(_textOutput) delete _textOutput;
	_textOutput = new TextUdfOutput(file,appendMode);
	return _textOutput->goodOutStream();
}
void BinaryInput::destroyTextOutput()
{
	if(_textOutput) delete _textOutput;
	_textOutput = 0;
}

/*
// Binary output.
bool BinaryInput::createBinaryOutput( const char* file, int divide_type )
{
	destroyBinaryOutput();
	bool appendMode = true;
	_binaryOutput = new BinaryOutput();
	switch(divide_type){
	case DCONST_COMBINE:
	case DCONST_NODIVIDE:
		if(_binaryOutput->open( file, appendMode )){
			return true;
		}
		break;
	case DCONST_DIVIDE:
		{
			appendMode = false;
			std::string directory;
			futl_replace_extension(file, DIVIDE_BINDIR_EXT, directory);
			if(directory.length() == 0){
				return false;
			}
			int dirf = fcache_directory(directory.c_str());
			if(dirf == -1){
				return false;
			}
			else if(dirf == 0){
				if(fcache_dir_create(directory.c_str())){
					return false;
				}
			}
			else if(dirf == 1){ // Read only.
				return false;
			}
			_binaryOutput->setDirectory(directory.c_str());
			std::string filepart;
			std::string basename;
			std::string extension;
			if(futl_filepart(file, filepart)){
				futl_split_extension(filepart, basename, extension);
				_binaryOutput->setRecordName(basename.c_str());
			}
			return true;
		}
		break;
	}
	return false;
}
void BinaryInput::destroyBinaryOutput()
{
	if(_binaryOutput) delete _binaryOutput;
	_binaryOutput = 0;
}
*/

// Divide record
void BinaryInput::checkBinaryDivide()
{
	if(!divide_record_data){
		divide_record_data = new BinaryDivideRecordFile();
	}
}
void BinaryInput::setDirectory(const char* dirname)
{
	checkBinaryDivide();
	divide_record_data->setDirectory(dirname);
}
const char* BinaryInput::getDirectory()
{
	checkBinaryDivide();
	return divide_record_data->getDirectory();
}
void BinaryInput::setRecordName(const char* recname)
{
	checkBinaryDivide();
	divide_record_data->setRecordName(recname);
}
const char* BinaryInput::getRecordName()
{
	checkBinaryDivide();
	return divide_record_data->getRecordName();
}
void BinaryInput::setRecordPath(const char* path)
{
	checkBinaryDivide();
	divide_record_data->setRecordPath(path);
}
const char* BinaryInput::getRecordPath()
{
	checkBinaryDivide();
	return divide_record_data->getRecordPath();
}
void BinaryInput::setRecordNumber(longint_t num)
{
	checkBinaryDivide();
	divide_record_data->setRecordNumber(num);
}
longint_t BinaryInput::getRecordNumber()
{
	checkBinaryDivide();
	return divide_record_data->getRecordNumber();
}

bool BinaryInput::initialReadProcess(
	FileStreamManager* fmgr,
	StructureCacheOperation* operation,
	int&         section,
	std::string& label,
	longint_t&   datasize,
	int*         shiftFineTuning,
	FILE*        binaryOpened
)
{
	if(getRecorder()){
		// With an additional process.(binary <---> text conversion.)
		return convertBinaryData(operation,section,label,datasize,shiftFineTuning,binaryOpened);
	}
	else{
		// Read only process.
		bool rb = operation->skipBinaryData(fmgr,section,label,datasize,shiftFineTuning,binaryOpened);
		if (rb) {
			std::string text;
			fmgr->getDivideDirectory(text);
			if (text.length() > 0) {
				operation->setDivideDirectory(text.c_str());
			}
		}
		return rb;
	}
}

/*
static bool getDividePath(
	string& divide_dir_path,
	string& divide_record_prifix,
	FileStreamManager* fmgr,
	StructureCacheOperation* operation
)
{
	bool directory_ok = false;
	if(operation->getDivideDirectory()){
		// Read from DIVD header data.
		divide_dir_path = operation->getDivideDirectory();
		int dir_status = fcache_directory(divide_dir_path.c_str());
		if(dir_status > 0){
			directory_ok = true;
		}
	}
	if(!directory_ok){
		// Use default data.
		string read_path = fmgr->getFile();
		if(divideBinaryDirName(read_path, divide_dir_path)){
			int dir_status = fcache_directory(divide_dir_path.c_str());
			if(dir_status > 0){
				directory_ok = true;
			}
		}
	}
	if(directory_ok){
		if(operation->getDivideRecordPrefix()){
			// Read from DIVD header data.
			divide_record_prifix = operation->getDivideRecordPrefix();
		} else {
			// Use default data.
			string filepart;
			string read_path = fmgr->getFile();
			if(futl_filepart(read_path.c_str(), filepart)){
				string extension;
				futl_split_extension(filepart, divide_record_prifix, extension);
			}
		}
	}
	return divide_record_prifix.length() > 0;
}
*/

// With an additional process.(binary <---> text converter.)
bool BinaryInput::convertBinaryData(
	StructureCacheOperation* operation,
	int&         section,
	std::string& label,
	longint_t&   datasize,
	int*         shiftFineTuning,
	FILE*        binaryOpened
)
{
	AbstractReadUDFRecord* recorder = getRecorder();
	if(!recorder) return false;
	MasterCacheMaker* maker = recorder->getMaker();
	FileStreamManager* fmgr = maker->getFileStream();
	DefineCache* define = maker->getDefine();
	// Text part.(Write definition)
	if(!recorder->operationDefine()) return false;
	// Read or copy.
	int convertFlag = 1; // 1:Text,-1:Binary (default:text).
	// Comparison of record numbers.
	unsigned int text_recordNumber = fmgr->getFileRecSize();
	if(text_recordNumber <= 0){
		if(recorder->operatedCommon()){
			if(recorder->getOperationRecordNumber() <= 0){
				// Output binary common data.
				if(!recorder->operationCommon()) return false;
			}
			text_recordNumber++;
		}
	}
	else{
		text_recordNumber++;
	}
	if(text_recordNumber >= recorder->getOperationRecordNumber()){
		convertFlag = -1; // binary.
	}
	switch(recorder->getConvertType()){
	case DCONST_CONVERT_BINARY:
	case DCONST_CONVERT_NO:
		convertFlag = -1;
		break;
	case DCONST_CONVERT_TEXT:
		break;
	}
	
	// The recorder stream is closed.
	//-if(recorder->alive()) recorder->terminate();
	//////////////////////////////////////////////////////////////
	if(convertFlag > 0){ // Output text udf.
		if(!fmgr->isDividedRecord()){
			if(!operation->readBinaryData(section,label,datasize,shiftFineTuning,0,binaryOpened)){
				return false;
			}
			if(datasize <= 0 && section != BINARY_READ_DIVIDE) return true; // End of file.
		} else {
			section = BINARY_READ_DIVIDE;
		}
		bool binary_divide_end_record = false;
		if(section == BINARY_READ_DIVIDE){
			if(!readDevideBinary(section, label, datasize, shiftFineTuning, fmgr, operation)){
				return false;
			}
			if(getRecordNumber() <= binary_RecordNumber){
				binary_divide_end_record = true;
			}
		}
		// Convert type [NORMAL RECORD/DIVIDE RECORD].
		short divide_record_mode = 0;
		switch(recorder->getDivideType()){
		case DCONST_DIVIDE:
			divide_record_mode = 2;
			break;
		default:
			divide_record_mode = 1;
		}
		// Record label.
		const char* labelp = label.c_str();
		// Record number.
		if(section <= 0){
			text_recordNumber = 0U;
		}
		else{
			if(text_recordNumber <= 0){
				text_recordNumber = 1;
			}
		}
		// Text writer.
		if(!_textOutput){
			if(!createTextOutput(recorder->getStringData(),true)) return false;
		}
		if(binary_divide_end_record){
			section = BINARY_READ_DIVIDE_END;
		}
		if(_textOutput->outputTextRecord(
			text_recordNumber,
			labelp,
			fmgr,
			define,
			operation,
			NULL, // family_name.
			NULL,
			divide_record_mode)){
			return recorder->operationNextRecord();
		}
	}
	else{ // Copy binary.
		bool binary_operation_flag = false;
		destroyTextOutput();
		BinaryOutput* binary_output = ((ConvertBinaryReadUDFRecord*)recorder)->getBinaryOutput();
		if(!recorder->operationFlag()){  // Already written out the binary header.
			binary_operation_flag = true;
			recorder->setOperationFlag(true);
			longint_t storePosition = tell();
			longint_t binaryPosition = fmgr->beginBinaryPosition();
			if(!seek( binaryPosition )) return false;
			// begin{binary}
			if(!binary_output->outBEGIN()) return false;
			// header.
			if(!operation->binaryCopy(binary_output->getOutput(), binaryOpened)) return false;
			// binary name map.
			if(!operation->binaryCopy(binary_output->getOutput(), binaryOpened)) return false;
			// write the divide record key.
			if(fmgr->getSaveDivideMode()){
				binary_output->outDIVIDE_HEADER(operation, 0);
			}
			// back to the record position.
			if(!seek( storePosition )) return false;
			// Check 'DIVD' key.
			int key = operation->readDivideKey(0, binaryOpened);
			if(key == -1){ /* DIVIDE RECORD */
				fmgr->setDividedRecord(true);
				// Set-up divide record file name.
				string divide_dir_path;
				string divide_rec_prifix;
				bool directory_ok = makeDivideBinaryPaths(
					divide_dir_path, divide_rec_prifix, fmgr, operation);
				if(directory_ok){
					setDirectory(divide_dir_path.c_str());
					setRecordName(divide_rec_prifix.c_str());
					// Search record.
					vector<string> file_names;
					size_t num = getRecordFiles(divide_dir_path, divide_rec_prifix, file_names, false);
					sort_names(file_names);
					longint_t max_divide_record = getMaxDivideRecordNumber(file_names);
					setRecordNumber(max_divide_record);
					// Copy commmon rcord data.
					BinaryInput readTemp;
					string rec_file_path;
					makeDivideRecordFileName(-1, divide_dir_path, divide_rec_prifix, rec_file_path);
					FILE* readOpened = readTemp.open(rec_file_path.c_str());
					if(readOpened){
						binary_output->copyRecord(
							0, operation, readOpened, &section, &label, &datasize);
					}
				} else {
					return false;
				}
			} else if(key == 0){ /* COMMON RECORD */
				 binary_RecordNumber = 0;
			} else if(key == 1){ /* STEP RECORD */
				binary_RecordNumber = 1;
			} else if(key == -9){
				// EOF
			} else {
				// ERROR
			}
			if(!seek( storePosition )) return false;
		}
		FILE* readOpened = binaryOpened;
		BinaryInput readTemp;
		bool binary_divide_end_record = false;
		if(fmgr->isDividedRecord()){ /* For reading file */
			binary_operation_flag = true;
			string dirpath = getDirectory();
			string record = getRecordName();
			string rec_file_path;
			makeDivideRecordFileName(binary_RecordNumber, dirpath, record, rec_file_path);
			readOpened = readTemp.open(rec_file_path.c_str());
			if(!readOpened){
				// Insert empty record data.
				BinaryDataWrite writer(binary_output->getOutput(), 0,0,0);
				writer.writeEmptyStep( 1, 0 );
				section = 1; // Count-up.
				return true; // SKIP COPYING THE RECORD.
			}
			if(getRecordNumber() <= binary_RecordNumber){
				binary_divide_end_record = true;
			}
		}
		// The first record number is checked from the binary record key.
		if(!binary_operation_flag && fmgr->isDividedRecord() && binary_RecordNumber <= 0){
			longint_t storePosition = ftell(readOpened);
			int keyid = BinaryDataRead::searchKey( readOpened );
			switch(keyid){
			case BINARY_DATA_START_KEYID:
			case BINARY_DATA_NAMEP_KEYID:
				return false;
			case BINARY_DATA_STEP0_KEYID:
				binary_RecordNumber = 0;
				break;
			case BINARY_DATA_STEPX_KEYID:
				binary_RecordNumber = 1;
				break;
			case BINARY_DATA_DIVID_KEYID:
			case BINARY_DATA_EOF_KEYID:
			default:
				return false;
			}
			if(!BinaryInput::seek( readOpened, storePosition )) return false;
		}
		bool rb = binary_output->copyRecord(
					binary_RecordNumber, operation, readOpened, &section, &label, &datasize);
		if(binary_divide_end_record){
			section = BINARY_READ_DIVIDE_END;
		}
		if(datasize == 0 && fmgr->getSaveDivideMode()){
			binary_output->setFileToDelete(true);
		}
		return rb;
	}
	return false;
}

// Read.
bool BinaryInput::readDevideBinary(
	int&         section,
	std::string& label,
	longint_t&   datasize,
	int*         shiftFineTuning,
	FileStreamManager*       fmgr,
	StructureCacheOperation* operation
)
{
	string divide_dir_path;
	string divide_rec_prifix;
	if(!fmgr->isDividedRecord()){
		fmgr->setDividedRecord(true);
		// Set-up divide record file name.
		bool directory_ok = makeDivideBinaryPaths(
			divide_dir_path, divide_rec_prifix, fmgr, operation);
		if(directory_ok){
			setDirectory(divide_dir_path.c_str());
			setRecordName(divide_rec_prifix.c_str());
		} else {
			return false;
		}
		// Search maximum record.
		vector<string> file_names;
		size_t num = getRecordFiles(divide_dir_path, divide_rec_prifix, file_names, false);
		sort_names(file_names);
		longint_t max_divide_record = getMaxDivideRecordNumber(file_names);
		setRecordNumber(max_divide_record);
		// Read commmon.
		BinaryInput readTemp;
		string rec_file_path;
		section = 0;
		makeDivideRecordFileName(-1, divide_dir_path, divide_rec_prifix, rec_file_path);
		FILE* readOpened = readTemp.open(rec_file_path.c_str());
		if(readOpened){
			if(!operation->readBinaryData(section,label,datasize,shiftFineTuning,0,readOpened)){
				return false;
			}
			binary_RecordNumber = -1;
		} else {
			// Read step.
			section = 1;
			makeDivideRecordFileName(binary_RecordNumber, divide_dir_path, divide_rec_prifix, rec_file_path);
			readOpened = readTemp.open(rec_file_path.c_str());
			if(readOpened){
				if(!operation->readBinaryData(section,label,datasize,shiftFineTuning,0,readOpened)){
					return false;
				}
			} else {
				// Empty or end record.
			}
		}
	} else {
		divide_dir_path = getDirectory();
		divide_rec_prifix = getRecordName();
		BinaryInput readTemp;
		string rec_file_path;
		section = 1;
		makeDivideRecordFileName(binary_RecordNumber, divide_dir_path, divide_rec_prifix, rec_file_path);
		FILE* readOpened = readTemp.open(rec_file_path.c_str());
		if(readOpened){
			if(!operation->readBinaryData(section,label,datasize,shiftFineTuning,0,readOpened)){
				return false;
			}
		} else {
			// Empty or end record.
		}
	}
	return true;
}

/*
 -99: binary header error.
 -999: record position error
 -1: fatal error( no record )
 0: normal end
 1: file or directory is read-only
 ---record error---
 10:file error
 11:some record has error
 ---divide binary udf---
 20: record directory exists but normal file
 21: record directory does not exist
*/
int BinaryInput::readBinaryInitial(
	unsigned int             lastRecord,
	FileStreamManager*       fileMgr,
	StructureCacheOperation* operation,
	unsigned int&            totalRecordNumber,
	const char*              serchLabel,
	bool                     readOnlyHeader
)
{
	int section;
	int shiftFineTuning;
	string label;
	unsigned int headerSize = 0;
	longint_t datasize = 0;
	// Position of 'begin{binary}'
	//longint_t currentPos = tell();
	longint_t currentPos = fileMgr->beginBinaryPosition();
	if(currentPos <= 0) return -999;  // record position error.
	if(!seek(currentPos)) return -999;
	// Read the binary header.
	if(operation->readBinaryHeader(NULL,&shiftFineTuning,_binaryOpend)){
		currentPos = currentPos + headerSize - shiftFineTuning;
	}
	else{
		return -99; // binary header error.
	}
	// Stop reading the following binary data.
	if(readOnlyHeader){
		if(ferror(_binaryOpend) == 0) return 0; // Normal(no error).
		return 10; // file error.
	}

	// Read records.
	totalRecordNumber = lastRecord;
	currentPos = tell(); /* Start record position. (COMN,STEP,DIVD) */
	binary_RecordNumber = lastRecord;
	while( feof(_binaryOpend) == 0 && ferror(_binaryOpend) == 0 ){
		// Read a binary record/common data.
		if(initialReadProcess(fileMgr,operation,section,label,datasize,&shiftFineTuning,_binaryOpend)){
			if(section == BINARY_READ_DIVIDE){
				// Divided binary record.
				fileMgr->setDividedRecord(true);
				/*
				-1: error
				20: exists but normal file
				21: not exist
				1: read only (warning)
				0: normal
				*/
				int rec_status = skipDividedRecords(fileMgr, operation, totalRecordNumber);
				return rec_status;
			}
			if(section == BINARY_READ_DIVIDE_END){ // Converter.
				return 0; // normal end.
			}
			if(datasize <= 0){
				if(getRecorder()){
					// Converter : Delete the last devided record file.
					BinaryOutput* binary_output = ((ConvertBinaryReadUDFRecord*)_read_recorder)->getBinaryOutput();
					binary_output->closeWrite();
					const char* out_file = binary_output->getRecordPath();
					if(out_file){ // The last devided record file is deleted.
						if(binary_output->isFileToDelete()){
							fcache_file_delete(out_file);
						}
						binary_output->setFileToDelete(false);
					}
					// End of file to delete.
					break;
				} else if(section < 0){
					break; // Normal binary end.
				}
			}
			// Initial(COMMON).
			if(section == 0){
				if(lastRecord > 0) return -9; // record specification error.
				fileMgr->setBinaryStartRecord(0U);
				fileMgr->setBinaryInitialRecPos(currentPos);
				existBinaryCommon = true;
			}
			// STEP.
			else if(section == 1){
				totalRecordNumber++;
				fileMgr->setRecInfo( totalRecordNumber, label.c_str(), 0, currentPos ); // Binary!
				// Serch label.
				if(serchLabel){
					if(label == serchLabel){
						_searchedRecord = totalRecordNumber;
					}
				}
			} else {
				return -1; // fatal error( no record ).
			}
			// Next.
			currentPos = currentPos + datasize - shiftFineTuning;
			binary_RecordNumber++;
		}
		else{
			if(binary_RecordNumber > 0) return 11; // some record has error.
			return -1; // fatal error( no record ).
		}
	}
	if(ferror(_binaryOpend) == 0) return 0; // Normal(no error).
	return 10; // file error.
}

bool BinaryInput::readBinaryRecord(
	BinaryFileInfo&          recordInfo,
	FileStreamManager*       fileMgr,
	StructureCacheOperation* operation
)
{
	bool selfOpened = false;
	bool retValue = false;
	if(!_binaryOpend){
		if(!recordInfo.fileName.length()) return false;
		if(!open(recordInfo.fileName.c_str())) return false;
		selfOpened = true;
	}
	// record info.
	REC_POSITION_TYPE currentPos = 0;
	if(recordInfo.recordNumber <= 0){
		currentPos = fileMgr->getBinaryInitialRecPos();
	}
	else{
		currentPos = recordInfo.recordPosition;
	}
	// Data positions
	SortedDataPositionList* data_position_list = fileMgr->getBinaryDataPosition(recordInfo.recordNumber);
	if(!data_position_list){
		data_position_list = new SortedDataPositionList();
		fileMgr->setBinaryDataPosition(recordInfo.recordNumber, data_position_list);
	}

	if(currentPos < 0) return false;
	if(!seek(currentPos)) return false;
	// Read step.
	int section;
	string label;
	longint_t datasize;
	int shiftFineTuning;
	retValue = operation->readBinaryData(section,label,datasize,&shiftFineTuning,data_position_list,_binaryOpend);
	if(selfOpened){
		if (retValue) {
			fileMgr->set_last_opened_file(recordInfo.fileName.c_str(), recordInfo.is_virtual_record);
		}
		close();
	}
	else {
		if (retValue) {
			fileMgr->set_last_opened_file(_binaryOpendFile.c_str(), _binaryOpendIsVirtual);
		}
	}
	return retValue;
}

bool BinaryInput::readBinaryRecord(
	unsigned int             recordNumber,
	FileStreamManager*       fileMgr,
	StructureCacheOperation* operation
)
{
	bool selfOpened = false;
	bool retValue = false;
	if(!_binaryOpend){
		char* fileName = fileMgr->getFile();
		if(!fileName) return false;
		if(!open(fileName)) return false;
		selfOpened = true;
	}
	// record info.
	REC_POSITION_TYPE currentPos = 0;
	if(recordNumber <= 0){
		currentPos = fileMgr->getBinaryInitialRecPos();
	}
	else{
		if(!fileMgr->getBinaryRecordInfo(recordNumber, currentPos, _get_virtual_info)) return false;
	}
	// Data positions
	SortedDataPositionList* data_position_list = fileMgr->getBinaryDataPosition(recordNumber);
	if(!data_position_list){
		data_position_list = new SortedDataPositionList();
		fileMgr->setBinaryDataPosition(recordNumber, data_position_list);
	}

	if(currentPos < 0) return false;
	if(!seek(currentPos)) return false;
	// Read step.
	int section;
	string label;
	longint_t datasize;
	int shiftFineTuning;
	retValue = operation->readBinaryData(section,label,datasize,&shiftFineTuning,data_position_list,_binaryOpend);
	if(selfOpened){
		close();
	}
	return retValue;
}

bool BinaryInput::rereadData(
	int		id,
	unsigned int             recordNumber,
	FileStreamManager*       fileMgr,
	StructureCacheOperation* operation
)
{
	bool selfOpened = false;
	bool retValue = false;
	if(!_binaryOpend){
		string fileName;
		if(!fileMgr->getReadFile(recordNumber, fileName)) return false;
		if(fileName.length() == 0) return false;
		if(!open(fileName.c_str())) return false;
		selfOpened = true;
	}
	// Data positions
	SortedDataPositionList* data_position_list = fileMgr->getBinaryDataPosition(recordNumber);
	// record info.
	REC_POSITION_TYPE currentPos = 0;
	if(recordNumber <= 0){
		currentPos = fileMgr->getBinaryInitialRecPos();
	}
	else{
		if(!fileMgr->getBinaryRecordInfo(recordNumber, currentPos, _get_virtual_info)) return false;
	}
	if(currentPos < 0) return false;
	if(!seek(currentPos)) return false;
	// Read step.
	retValue = operation->rereadBinaryData(id,data_position_list,_binaryOpend);
	if(selfOpened){
		close();
	}
	return retValue;
}

bool BinaryInput::rereadAllData(
	unsigned int             recordNumber,
	FileStreamManager*       fileMgr,
	StructureCacheOperation* operation
)
{
	bool selfOpened = false;
	bool retValue = false;
	if(!_binaryOpend){
		string fileName;
		if(!fileMgr->getReadFile(recordNumber, fileName)) return false;
		if(fileName.length() == 0) return false;
		if(!open(fileName.c_str())) return false;
		selfOpened = true;
	}
	// record info.
	REC_POSITION_TYPE currentPos = 0;
	if(recordNumber <= 0){
		currentPos = fileMgr->getBinaryInitialRecPos();
	}
	else{
		if(!fileMgr->getBinaryRecordInfo(recordNumber,currentPos, _get_virtual_info)) return false;
	}
	if(currentPos < 0) return false;
	if(!seek(currentPos)) return false;
	// Read step.
	retValue = operation->rereadBinaryAllData(_binaryOpend);
	if(selfOpened){
		close();
	}
	return retValue;
}

/*
 -1: error
 20: exists but normal file
 21: not exist
 1: read only (warning)
 0: normal
 */
int BinaryInput::skipDividedRecords(
	FileStreamManager*       fileMgr,
	StructureCacheOperation* operation,
	unsigned int&            totalRecordNumber
)
{
	// Directory name.
	string directory_name;
	char* read_file = fileMgr->getFile();
	string sourceName = read_file;
	// Read name.
	const char* dir_p = operation->getDivideDirectory();
	if(dir_p){
		directory_name = dir_p;
	} else {
		// default path
		if(!getDivideBinaryDirName(sourceName, directory_name)){
			return -1;
		}
	}

	int result = 0;
	int dir_status = fcache_directory(directory_name.c_str());
	switch (dir_status) {
	case -1: /* error:exists but normal file */
		return 20;
	case 0: /* not exist */
		return 21;
	case 1: /* read only (warning) */
		result = 1;
		break;
	case 2: /* writable (normal) */
		break;
	}

	// Record file name.
	string basename;
	const char* file_p = operation->getDivideRecordPrefix();
	if(file_p){
		basename = file_p;
	} else {
		// default name.
		if(!getDivideBinaryRecordPrefix(sourceName, basename)){
			return -1;
		}
	}

	// variables in loop.
	int section = 1;
	string label;
	longint_t datasize = 0;
	int shiftFineTuning = 0;

	// Get record files. 
	vector<string> file_names;
	size_t num = getRecordFiles(directory_name, basename, file_names, false);
	for(size_t i=0;i<file_names.size();i++){
		longint_t record_number = getDivideRecordNumber(file_names[i]);
		if(record_number <= -2){
			continue;
		}
		totalRecordNumber = record_number + 1;
		string path = directory_name + fcache_get_delimiter() + file_names[i];
		if(record_number == -1){
			// Read commmon data.
			BinaryInput binRead;
			FILE* binaryOpened = binRead.open(path.c_str());
			if(!binaryOpened){
				continue;
			}
			fileMgr->setBinaryStartRecord(0U);
			fileMgr->setBinaryInitialRecPos(0);
			//- fileMgr->setDividedRecord( true );
			section = 1;
			label.clear();
			datasize = 0;
			shiftFineTuning = 0;
			if(!operation->skipBinaryData(fileMgr,section,label,datasize,&shiftFineTuning,binaryOpened)){
				if(operation->getBinaryReadError()){
					// message
					return -1;
				}
				return -1;
			}
			fileMgr->setRecInfo( totalRecordNumber, 0, 0, 0, path.c_str(), true); // Divide Binary!
			existBinaryCommon = true;
		}
		else{
			// Step record file.
			BinaryInput binRead;
			FILE* binaryOpened = binRead.open(path.c_str());
			if(!binaryOpened){
				continue;
			}
			section = 1;
			label.clear();
			datasize = 0;
			shiftFineTuning = 0;
			if(!operation->skipBinaryData(fileMgr,section,label,datasize,&shiftFineTuning,binaryOpened)){
				return -1;
			}
			fileMgr->setRecInfo( totalRecordNumber, label.c_str(), 0, 0, path.c_str(), true); // Divide Binary!
		}
	}
	return result;
}

bool readDivideRecord(
	FileStreamManager* fileMgr
)
{
	BinaryInput binRead;
	if(!binRead.open(fileMgr->getFile())){
		return false;
	}
	return true;
}
